"""Introduction
~~~~~~~~~~~~~~~

The :mod:`libraw` package contains low-level CTYPES_ based APIs for interfacing
with LibRaw_ by LibRaw, LLC.

While this library can be used on its own to access the full functionality of
LibRaw and develop raw photos, we recommend using the higher-level :mod:`rawkit`
module, which provides a more pythonic interface to LibRaw.

.. _CTYPES: https://docs.python.org/3/library/ctypes.html
.. _LibRaw: http://www.libraw.org
"""
